import enum
import json
import os
from abc import ABC, abstractmethod
from typing import Any, Optional


class Storage(enum.Enum):
    Memory = 1,
    File = 2


class Cache(ABC):
    @abstractmethod
    def get(self, key: str) -> Any:
        pass

    @abstractmethod
    def set(self, key: str, value: Any) -> None:
        pass


class MemoryCache(Cache):
    cache = {}

    def get(self, key: int) -> Any:
        return self.cache.get(key)

    def set(self, key: int, value: Any) -> None:
        self.cache[key] = value


class FileCache(Cache):
    def get(self, key: int) -> Optional[Any]:
        filename = self._get_file_name(key)
        exists = os.path.exists(filename)

        if not exists:
            return None

        return json.load(open(filename))

    def set(self, key: int, value: Any) -> None:
        filename = self._get_file_name(key)
        f = open(filename, "x")
        json.dump(value, f)

    @staticmethod
    def _get_file_name(key):
        return f'{key}.cache'


cache_types = {
    Storage.Memory: MemoryCache,
    Storage.File: FileCache
}


class Cacher:
    def __init__(self, storage: Storage = Storage.Memory, on_hit_return_value: Optional[Any] = None, key_fn=None):
        self.key_fn = key_fn
        self.on_hit_return_value = on_hit_return_value
        self.cacher = cache_types[storage]()

    def __call__(self, func, *args, **kwargs):
        def wrapper(*a, **kw):
            key = self.key_fn() if self.key_fn else self._make_key(func.__name__, a, kw)
            result = self.cacher.get(key)

            if result:
                return result if self.on_hit_return_value is None else self.on_hit_return_value

            result = func(*a, **kw)
            self.cacher.set(key, result)
            return result

        wrapper.__name__ = func.__name__
        return wrapper

    @staticmethod
    def _make_key(name: str, args, kwargs) -> int:
        key = args
        if kwargs:
            for item in kwargs.items():
                key += item
        return hash((name,) + key)



import functools
import os
from dataclasses import dataclass
from datetime import datetime
from typing import Dict, Set, List, Optional

import monodiff.git_client
from monodiff.base_provider import MonoProvider
from monodiff.cache_decorator import Cacher, Storage
from monodiff.thread2 import ThreadWithReturnValue
from monodiff.time_decorator import Timer


@dataclass
class Affected:
    changed_files: List[str]
    affected_projects: Set[str]
    affected_apps: Set[str]


class Mono:
    def __init__(self, provider: MonoProvider):
        self.provider = provider

    def get_affected_projects(self, branch: Optional[str]) -> Affected:
        csproj_to_apps = self.build_projects_to_apps_mapping()
        changed_files: List[str] = git_client.get_diff_compared_to(branch)
        changed_projects: Set[str] = self._get_changed_projects(changed_files)
        dependant_apps: Set[str] = Mono._get_dependant_apps(changed_projects, csproj_to_apps)
        return Affected(changed_files=changed_files, affected_projects=changed_projects, affected_apps=dependant_apps)

    def get_app_projects(self) -> List[str]:
        projects = self.provider.find_projects(recursive=True)
        return [os.path.basename(project) for project in projects if self.provider.project_is_an_app(project)]

    def _get_proj_deps_recursive(self, dependencies: Set[str], project_mapping: Dict[str, str], cache_scope: str) -> Set[str]:
        if len(dependencies) == 0:
            return set()

        all_deps = set()
        for index, dependency in enumerate(dependencies):
            dep_deps = self.provider.get_dependant_projects(project_mapping[dependency], cache_scope)
            all_deps = all_deps.union(dep_deps)

            recursive_deps = self._get_proj_deps_recursive(dep_deps, project_mapping, cache_scope)
            all_deps = all_deps.union(recursive_deps)

        return all_deps

    @Cacher(storage=Storage.File, key_fn=git_client.get_last_commit_hash)
    @Timer()
    def build_projects_to_apps_mapping(self) -> Dict[str, List]:
        project_paths: List[str] = self.provider.find_projects(recursive=True)
        project_mappings = self._get_project_path_mappings(project_paths)
        apps = [project for project in project_mappings.keys() if self.provider.project_is_an_app(project_mappings[project])]
        app_to_threads_map = {app: self._get_dependencies_of_app_thread(app, project_mappings) for app in apps}
        app_to_dependencies_map = {app: thread.join() for app, thread in app_to_threads_map.items()}
        return self._flip(app_to_dependencies_map)

    @staticmethod
    def _get_project_path_mappings(project_paths: List[str]) -> Dict[str, str]:
        return {os.path.basename(project_path): project_path for project_path in project_paths}

    @staticmethod
    def _flip(app_to_dependencies_map: Dict[str, List[str]]) -> Dict[str, List[str]]:
        projects_to_apps: Dict[str, List[str]] = {}
        for app, app_dependencies in app_to_dependencies_map.items():
            for dep in app_dependencies:
                if dep not in projects_to_apps:
                    projects_to_apps[dep] = []

                projects_to_apps[dep].append(app)
        return projects_to_apps

    def _get_dependencies_of_app(self, app_project: str, project_mappings: Dict[str, str]) -> Set[str]:
        start = datetime.now()
        print('start', app_project)
        first_line_deps = self.provider.get_dependant_projects(project_mappings[app_project], app_project)
        dependencies = self._get_proj_deps_recursive(first_line_deps, project_mappings, app_project).union(first_line_deps)
        dependencies.add(app_project)
        print('finish', app_project, datetime.now() - start)
        return dependencies

    def _get_dependencies_of_app_thread(self, app_project: str, project_mappings: Dict[str, str]) -> ThreadWithReturnValue:
        thread = ThreadWithReturnValue(target=self._get_dependencies_of_app, args=(app_project, project_mappings))
        thread.start()
        return thread

    @Cacher()
    def _get_closest_csproj(self, path: str) -> Optional[str]:
        if not path:
            return None

        dir_path = os.path.dirname(path)
        projects = self.provider.find_projects(prefix=dir_path + '/', recursive=True)
        return os.path.basename(projects[0]) if any(projects) else self._get_closest_csproj(dir_path)

    def _get_changed_projects(self, changed: List[str]) -> Set[str]:
        return set(map(self._get_closest_csproj, changed))

    @staticmethod
    def _get_dependant_apps(projects_changed: Set[str], projects_to_apps: Dict[str, List[str]]) -> Set[str]:
        apps_changed_map = list(map(lambda project: set(projects_to_apps.get(project, [])), projects_changed))

        if not any(apps_changed_map):
            return set()

        return functools.reduce(lambda accumulated, current: (accumulated or set()).union(current), apps_changed_map)



from typing import List, Optional

import monodiff.bash_client


def get_diff_compared_to(branch: Optional[str]) -> List[str]:
    cmd = "git diff --name-only"

    if branch:
        cmd = f'{cmd} {branch}'

    return bash_client.execute_cmd(cmd)


def get_last_commit_hash() -> str:
    cmd = 'git rev-parse --short HEAD '
    return bash_client.execute_cmd(cmd)[0]



import glob
import os
from typing import Set, Optional, List
from tenacity import retry, wait_fixed, stop_after_attempt

import monodiff.bash_client
from monodiff.base_provider import MonoProvider
from monodiff.cache_decorator import Cacher


class DotnetProvider(MonoProvider):
    def project_is_an_app(self, project: str) -> bool:
        with open(project) as project_file:
            return '<AppName>' in project_file.read()

    # circular dependencies fix
    @Cacher(on_hit_return_value=set())
    @retry(reraise=True, wait=wait_fixed(0.3), stop=stop_after_attempt(3))
    def get_dependant_projects(self, project_name: str, cache_scope: str) -> Set[str]:
        cmd = f"dotnet list {project_name} reference"
        output_lines = bash_client.execute_cmd(cmd)
        return set([os.path.basename(line.replace('\\', '/')) for line in output_lines[2:]])

    def find_projects(self, prefix: str = '', recursive: Optional[bool] = False) -> List[str]:
        return glob.glob(f'{prefix}**/*.csproj', recursive=recursive)



from abc import abstractmethod, ABC
from typing import Optional, Set, List


class MonoProvider(ABC):
    @abstractmethod
    def project_is_an_app(self, project: str) -> bool:
        pass

    @abstractmethod
    def get_dependant_projects(self, project: str, cache_scope=str) -> Set[str]:
        pass

    @abstractmethod
    def find_projects(self, prefix: str = '', recursive: Optional[bool] = False) -> List[str]:
        pass


from typing import Optional
import fire

from monodiff.dotnet_provider import DotnetProvider
from monodiff.mono import Mono
import logging

class MonoCli:
    def __init__(self):
        self.logger = logging.getLogger("mono")
        self.mono = Mono(DotnetProvider())

    def affected(self, branch: Optional[str] = None):
        affected = self.mono.get_affected_projects(branch)
        print(f"*************************************************")
        print(f"*************************************************")
        print(f"Files Changed: {affected.changed_files}")
        print(f"Projects affected: {affected.affected_projects}")
        print(f"******* Please build the following apps")
        return affected.affected_apps

    def build(self):
        return self.mono.build_projects_to_apps_mapping()

    def apps(self):
        return self.mono.get_app_projects()

def main():
    fire.Fire(MonoCli)

if __name__ == '__main__':
    fire.Fire(MonoCli)



from datetime import datetime


class Timer:
    def __call__(self, func, *args, **kwargs):
        def wrapper(*a, **kw):
            start = datetime.now()
            result = func(*a, **kw)
            end = datetime.now()
            print(func.__name__, a, kw, end - start)
            return result

        wrapper.__name__ = func.__name__
        return wrapper

    @staticmethod
    def _make_key(name: str, args, kwargs) -> int:
        key = args
        if kwargs:
            for item in kwargs.items():
                key += item
        return hash((name,) + key)



from threading import Thread


class ThreadWithReturnValue(Thread):
    def __init__(self, group=None, target=None, name=None, args=(), kwargs={}):
        Thread.__init__(self, group, target, name, args, kwargs)
        self._return = None

    def run(self):
        if self._target is not None:
            self._return = self._target(*self._args, **self._kwargs)

    def join(self, timeout=None):
        Thread.join(self, timeout)
        return self._return
